package de.fhdw.gaming.ipspiel23.ht.domain.impl;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Pattern;

import de.fhdw.gaming.core.domain.GameBuilder;
import de.fhdw.gaming.core.domain.GameBuilderFactory;
import de.fhdw.gaming.core.domain.GameException;
import de.fhdw.gaming.core.domain.Strategy;
import de.fhdw.gaming.core.ui.InputProvider;
import de.fhdw.gaming.core.ui.InputProviderException;
import de.fhdw.gaming.core.ui.type.validator.MaxValueValidator;
import de.fhdw.gaming.core.ui.type.validator.MinValueValidator;
import de.fhdw.gaming.core.ui.type.validator.PatternValidator;
import de.fhdw.gaming.core.ui.type.validator.ValueSetValidator;
import de.fhdw.gaming.ipspiel23.ht.domain.Answer;
import de.fhdw.gaming.ipspiel23.ht.domain.IHTGameBuilder;
import de.fhdw.gaming.ipspiel23.ht.domain.IHTGameBuilderFactory;
import de.fhdw.gaming.ipspiel23.ht.domain.IHTPlayer;
import de.fhdw.gaming.ipspiel23.ht.domain.IHTPlayerBuilder;
import de.fhdw.gaming.ipspiel23.ht.moves.factory.HTDefaultMoveFactory;
import de.fhdw.gaming.ipspiel23.ht.moves.factory.IHTMoveFactory;
import de.fhdw.gaming.ipspiel23.ht.strategy.IHTStrategy;
import de.fhdw.gaming.ipspiel23.ht.strategy.IHTStrategyFactory;
import de.fhdw.gaming.ipspiel23.ht.strategy.factory.HTDefaultStrategyFactoryProvider;
import de.fhdw.gaming.ipspiel23.ht.strategy.factory.IHTStrategyFactoryProvider;

/**
 * Represents a factory for game builder instances.
 */
public final class HTGameBuilderFactory implements IHTGameBuilderFactory {

    /**
     * The number of players.
     */
    private static final int NUMBER_OF_PLAYERS = 2;

    /**
     * Smallest allowed maximum computation time per move in seconds.
     */
    private static final int MIN_MAX_COMPUTATION_TIME_PER_MOVE = 1;
    
    /**
     * Largest allowed maximum computation time per move in seconds.
     */
    private static final int MAX_MAX_COMPUTATION_TIME_PER_MOVE = 3600;

    /**
     * The available strategies.
     */
    private final Set<IHTStrategy> strategies;

    /**
     * Creates a new instance of {@link HTGameBuilderFactory}.
     */
    public HTGameBuilderFactory() {
        this(new HTDefaultStrategyFactoryProvider());
    }

    /**
     * Creates a new instance of {@link HTGameBuilderFactory}.
     *
     * @param provider The provider for the strategy factories.
     */
    public HTGameBuilderFactory(final IHTStrategyFactoryProvider provider) {
        final IHTMoveFactory moveFactory = new HTDefaultMoveFactory();

        final List<IHTStrategyFactory> factories = provider.getStrategyFactories();
        this.strategies = new LinkedHashSet<>();
        for (final IHTStrategyFactory factory : factories) {
            this.strategies.add(factory.create(moveFactory));
        }
    }

    @Override
    public String getName() {
        return "Heads or Tails";
    }

    @Override
    public int getMinimumNumberOfPlayers() {
        return NUMBER_OF_PLAYERS;
    }

    @Override
    public int getMaximumNumberOfPlayers() {
        return NUMBER_OF_PLAYERS;
    }

    @Override
    public List<? extends Strategy<?, ?, ?>> getStrategies() {
        return new ArrayList<>(this.strategies);
    }

    @Override
    public IHTGameBuilder createGameBuilder(final InputProvider inputProvider) throws GameException {
        try {
            final IHTGameBuilder gameBuilder = new HTGameBuilder();

            @SuppressWarnings("unchecked")
            final Map<String,
                Object> gameData = inputProvider.needInteger(
                    GameBuilderFactory.PARAM_MAX_COMPUTATION_TIME_PER_MOVE,
                    "Maximum computation time per move in seconds",
                    Optional.of(GameBuilder.DEFAULT_MAX_COMPUTATION_TIME_PER_MOVE),
                    new MinValueValidator<>(HTGameBuilderFactory.MIN_MAX_COMPUTATION_TIME_PER_MOVE),
                    new MaxValueValidator<>(HTGameBuilderFactory.MAX_MAX_COMPUTATION_TIME_PER_MOVE))
                    .requestData("Game properties");

            gameBuilder.changeMaximumComputationTimePerMove(
                    (Integer) gameData.get(GameBuilderFactory.PARAM_MAX_COMPUTATION_TIME_PER_MOVE));

            final InputProvider playerInputProvider1 = inputProvider.getNext(gameData);
            final Map<String, Object> playerData1 = this.requestPlayerData(playerInputProvider1, "Player 1", 1);
            final IHTPlayer player1 = this.createPlayer(gameBuilder.createPlayerBuilder(), playerData1);
            final IHTStrategy playerStrategy1 = this.getStrategy(playerData1);
            gameBuilder.addPlayer(player1, playerStrategy1);

            final InputProvider playerInputProvider2 = playerInputProvider1.getNext(playerData1);
            final Map<String, Object> playerData2 = this.requestPlayerData(playerInputProvider2, "Player 2", -1);
            final IHTPlayer player2 = this.createPlayer(gameBuilder.createPlayerBuilder(), playerData2);
            final IHTStrategy playerStrategy2 = this.getStrategy(playerData2);
            gameBuilder.addPlayer(player2, playerStrategy2);

            if (player1.getPossibleOutcomes().equals(player2.getPossibleOutcomes())) {
                throw new GameException("Players cannot bet on the same outcome!");
            }
            
            return gameBuilder;
        } catch (final InputProviderException e) {
            throw new GameException(String.format("Creating HT game was aborted: %s", e.getMessage()), e);
        }
    }

    /**
     * Returns data for a player builder.
     *
     * @param inputProvider The input provider.
     * @param title         The title for the UI.
     * @param defaultValueFactor   The factor to be multiplied with the outcomes (allows easily inverting outcomes
     * /pre-filling input provider stuff)
     * @throws InputProviderException if the operation has been aborted prematurely (e.g. if the user cancelled a
     *                                dialog).
     */
    @SuppressWarnings("unchecked")
    private Map<String, Object> requestPlayerData(final InputProvider inputProvider, final String title, 
            final int defaultValueFactor)
            throws GameException, InputProviderException {
        final Set<Integer> allowedOutcomeValues = new HashSet<>();
        allowedOutcomeValues.add(1);
        allowedOutcomeValues.add(-1);
        
        final ValueSetValidator<Integer> outcomeValidator = new ValueSetValidator<>(allowedOutcomeValues);
        
        inputProvider
            .needString(
                GameBuilderFactory.PARAM_PLAYER_NAME,
                "Name",
                Optional.empty(),
                new PatternValidator(Pattern.compile("\\S+(\\s+\\S+)*")))
            .needInteger(
                PARAM_PLAYER_OUTCOME_EQUAL,
                "Player's outcome on equal sides (HH/TT/EE)",
                Optional.of(1 * defaultValueFactor),
                outcomeValidator)
            .needInteger(
                PARAM_PLAYER_OUTCOME_DIFFERENT,
                "Player's outcome on different sides (HT/HE/TH/TE/EH/ET)",
                Optional.of(-1 * defaultValueFactor),
                outcomeValidator)
            .needObject(GameBuilderFactory.PARAM_PLAYER_STRATEGY, "Strategy", Optional.empty(), this.strategies);

        return inputProvider.requestData(title);
    }

    /**
     * Creates a HT player.
     *
     * @param playerBuilder The player builder.
     * @param playerData    The requested player data.
     * @return The created {@link IHTPlayer}.
     * @throws InputProviderException if the operation has been aborted prematurely (e.g. if the user cancelled a
     *                                dialog).
     */
    private IHTPlayer createPlayer(final IHTPlayerBuilder playerBuilder,
            final Map<String, Object> playerData) throws GameException, InputProviderException {
        final double equalOutcome = (double) (Integer) playerData.get(PARAM_PLAYER_OUTCOME_EQUAL);
        final double differentOutcome = (double) (Integer) playerData.get(PARAM_PLAYER_OUTCOME_DIFFERENT);
        
        if (equalOutcome == differentOutcome) {
            throw new GameException("Outcomes for equal and different sides must not be identical.");
        }
        
        final Map<Answer, Map<Answer, Double>> possibleOutcomes = new LinkedHashMap<>();
        
        final Map<Answer, Double> possibleOutcomesHeads = new LinkedHashMap<>();
        possibleOutcomesHeads.put(Answer.HEADS, equalOutcome);
        possibleOutcomesHeads.put(Answer.TAILS, differentOutcome);
        possibleOutcomesHeads.put(Answer.EDGE, differentOutcome);
        possibleOutcomes.put(Answer.HEADS, possibleOutcomesHeads);

        final Map<Answer, Double> possibleOutcomesTails = new LinkedHashMap<>();
        possibleOutcomesTails.put(Answer.HEADS, differentOutcome);
        possibleOutcomesTails.put(Answer.TAILS, equalOutcome);
        possibleOutcomesTails.put(Answer.EDGE, differentOutcome);
        possibleOutcomes.put(Answer.TAILS, possibleOutcomesTails);
        
        final Map<Answer, Double> possibleOutcomesEdge = new LinkedHashMap<>();
        possibleOutcomesEdge.put(Answer.HEADS, differentOutcome);
        possibleOutcomesEdge.put(Answer.TAILS, differentOutcome);
        possibleOutcomesEdge.put(Answer.EDGE, equalOutcome);
        possibleOutcomes.put(Answer.EDGE, possibleOutcomesEdge);

        return playerBuilder.changeName((String) playerData.get(GameBuilderFactory.PARAM_PLAYER_NAME))
            .changePossibleOutcomes(possibleOutcomes).build();
    }

    /**
     * Returns a HT strategy.
     *
     * @param playerData The requested player data.
     * @return The HT strategy.
     */
    private IHTStrategy getStrategy(final Map<String, Object> playerData) {
        return (IHTStrategy) playerData.get(GameBuilderFactory.PARAM_PLAYER_STRATEGY);
    }
}
